/*
 * Copyright 2014-present Facebook, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.facebook.litho.specmodels.generator;

import static com.facebook.litho.specmodels.generator.ComponentBodyGenerator.getStateContainerClassName;
import static com.facebook.litho.specmodels.generator.GeneratorConstants.STATE_CONTAINER_FIELD_NAME;

import com.facebook.litho.specmodels.model.SpecModel;
import com.squareup.javapoet.ClassName;
import com.squareup.javapoet.MethodSpec;
import javax.lang.model.element.Modifier;

/**
 * Class that generates the preamble for a Component.
 */
public class PreambleGenerator {

  private PreambleGenerator() {
  }

  public static TypeSpecDataHolder generate(SpecModel specModel) {
    return TypeSpecDataHolder.newBuilder()
        .addTypeSpecDataHolder(generateSourceDelegate(specModel))
        .addTypeSpecDataHolder(generateConstructor(specModel))
        .build();
  }

  /**
   * Generate a delegate to the Spec that defines this component.
   */
  static TypeSpecDataHolder generateSourceDelegate(SpecModel specModel) {
    final TypeSpecDataHolder.Builder sourceDelegateBuilder = TypeSpecDataHolder.newBuilder();

    if (specModel.hasInjectedDependencies()
        && specModel.getDependencyInjectionHelper().hasSpecInjection()) {
      sourceDelegateBuilder.addTypeSpecDataHolder(
          specModel.getDependencyInjectionHelper().generateSourceDelegate(specModel));
    }

    return sourceDelegateBuilder.build();
  }

  /**
   * If the spec has injected dependencies, generate a DI constructor. Otherwise, generate a
   * private constructor to enforce singleton-ity.
   */
  static TypeSpecDataHolder generateConstructor(SpecModel specModel) {
    final MethodSpec.Builder constructorBuilder =
        MethodSpec.constructorBuilder().addStatement("super($S)", specModel.getComponentName());

    if (specModel.hasInjectedDependencies()) {
      final MethodSpec diConstructor =
          specModel.getDependencyInjectionHelper().generateConstructor(specModel);

      constructorBuilder
          .addAnnotations(diConstructor.annotations)
          .addCode(diConstructor.code)
          .addModifiers(diConstructor.modifiers)
          .addParameters(diConstructor.parameters);
    } else {
      constructorBuilder.addModifiers(Modifier.PRIVATE);
    }

    final boolean hasState = !specModel.getStateValues().isEmpty();
    if (hasState) {
      final ClassName stateContainerClass =
          ClassName.bestGuess(getStateContainerClassName(specModel));
      constructorBuilder.addStatement(
          STATE_CONTAINER_FIELD_NAME + " = new $T()", stateContainerClass);
    }

    return TypeSpecDataHolder.newBuilder().addMethod(constructorBuilder.build()).build();
  }
}
